import asyncio

from .protocol import DnsTCP


class Connection:

    def __init__(self, dispatcher_cls, loop, host='0.0.0.0', port=53):
        self.dispatcher_cls = dispatcher_cls
        self.loop = loop
        self.host = host
        self.port = port

    async def start(self):
        # Get a reference to the event loop as we plan to use
        # low-level APIs.

        server = await self.loop.create_server(
            lambda: DnsTCP(self, self.loop),
            self.host, self.port)

        async with server:
            await server.serve_forever()



import asyncio
from typing import Union, Text, Tuple


class DnsUDP(asyncio.DatagramProtocol):

    def __init__(self, connection, loop, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loop = loop
        self.connection = connection

    def connection_made(self, transport: asyncio.transports.BaseTransport) -> None:
        self.transport = transport

    def datagram_received(self, data: Union[bytes, Text], addr: Tuple[str, int]) -> None:
        dispatcher = self.connection.dispatcher_cls(self.loop)
        task = asyncio.Task(self.to_async(dispatcher, data, addr))

    @asyncio.coroutine
    def to_async(self, dispatcher, data, addr):
        result_data = yield from dispatcher.handle(data)
        self.transport.sendto(result_data, addr)






type_int = {
    1: 'a',
    28: 'aaaa',
    15: 'mx'
}


class Resolver:

    def __init__(self, connection):
        self.connection = connection

    async def find(self, type, cls, name) -> dict:
        # TODO
        domain = self.connection.dictionary.get(name)
        if domain:
           for rr in domain.get(type_int[type]):
               print(rr)
        return {}


import asyncio

from .protocol import DnsUDP


class Connection:

    def __init__(self, dispatcher_cls, loop, host='0.0.0.0', port=53):
        self.dispatcher_cls = dispatcher_cls
        self.loop = loop
        self.host = host
        self.port = port

    async def start(self):
        print("Starting UDP server")
        transport, protocol = await self.loop.create_datagram_endpoint(
            lambda: DnsUDP(self, self.loop),
            local_addr=('0.0.0.0', 53))
        try:
            await asyncio.sleep(99999999999999)
        finally:
            transport.close()



from neptune import config


class Connector:

    def __init__(self, loop):
        if not hasattr(self, 'connected'):
            self.connected = False
            self.loop = loop
            self.resolve_connectors = []
            self.cache_connector = None
            self.loop.create_task(self.init_connections())

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            instance = super().__new__(cls)
            cls.instance = instance
        return cls.instance

    async def init_connections(self):
        if not self.connected:
            for resolver in config.RESOLVERS:
                setattr(self, resolver, __import__(resolver, globals=globals(), fromlist=['']))
                con = getattr(self, resolver).Connection(loop=self.loop, server_connector=self,
                                                         **getattr(config, resolver.replace('.', '_').upper(), {}))
                self.resolve_connectors.append(con)
            if hasattr(config, 'CACHE'):
                setattr(self, config.CACHE, __import__(config.CACHE, globals=globals()))
                con = getattr(self, config.CACHE).Connection(loop=self.loop,
                                                             **getattr(config, config.CACHE.upper(), {}))
                await con.init_connection()
                self.cache_connector = con
        self.connected = True

    async def resolve(self, name, type, cls):
        for con in self.resolve_connectors:
            result = await con.Resolver.find(type=type,
                                             cls=cls,
                                             name=name)
            if result:
                return result
        return None

    async def cache_set(self, name, type, cls, data, expire):
        assert hasattr(self, 'cache_connector')
        return self.cache_connector.Cache.set(name, type, cls, data, expire)

    async def cache_get(self, name, type, cls):
        assert hasattr(self, 'cache_connector')
        return self.cache_connector.Cache.get(name, type, cls)



import asyncio
loop = asyncio.new_event_loop()
# Priority order kept

RESOLVERS = [
    # 'neptune.resolver.dictionary',
    # 'neptune_resolver_default',
    'triton.resolver'
    # 'neptune_resolver_rest',
    # 'neptune_resolver_postgres'
]

NEPTUNE_RESOLVER_DICTIONARY = {
    'dictionary': {
        'example.com': {
            'a': {'127.0.0.1', '127.0.0.2'}
        }
    }
}

CACHE = 'neptune_cache_redis'


NEPTUNE_CACHE_REDIS = {
    'host': 'localhost',
    'database': 1
}

NEPTUNE_RESOLVER_REST = {
    'base_url': '',
    'username': '',
    'password': '',
}

NEPTUNE_RESOLVER_POSTGRES = {

}

PROTOCOLS = [
    'neptune.protocols.udp',
    'neptune.protocols.tcp',
    'neptune_dnsoverhttps_protocol'
]

NEPTUNE_DNSOVERHTTPS_PROTOCOL = {
    'prefix': 'dns-request',
    'cert_path': '',
    'key_path': '',
    'host': '0.0.0.0',
    'port': '8090',
    'with_ssl': False
}

NEPTUNE_PROTOCOLS_UDP = {
    'host': '0.0.0.0',
    'port': 53
}

NEPTUNE_PROTOCOLS_TCP = {
    'host': '0.0.0.0',
    'port': 53
}

NEPTUNE_RESOLVER_DEFAULT = {
    'resolvers': ['1.1.1.1', '8.8.8.8'] # Your favorite DNS resolvers
}



from neptune import config
from neptune.protocols.dispatcher import Dispatcher


class Server:
    def __init__(self, loop):
        self.loop = loop
        self.protocols = []

    def load_protocols(self):
        for proto in config.PROTOCOLS:
            _proto = __import__(proto, fromlist=[''])
            con = _proto.Connection(**{**getattr(config, proto.replace('.', '_').upper()),
                                       'loop': self.loop, 'dispatcher_cls': Dispatcher})
            self.protocols.append(con)

    async def start(self):
        self.load_protocols()
        for proto in self.protocols:
            self.loop.create_task(proto.start())


if __name__ == '__main__':
    server = Server(config.loop)
    config.loop.run_until_complete(server.start())
    config.loop.run_forever()



# from .._extentions.dnssec import sign_rrset
from neptune.connector import Connector
from triton.dns.edns import Edns
from triton.dns.message import Message
import json


class Dispatcher:
    able_to_truncate = False

    def __init__(self, loop, can_truncate=False):
        self.connector = Connector(loop)
        self.loop = loop
        self.can_truncate = can_truncate

    async def handle(self, raw_data):
        message = self.build_message(raw_data)
        Edns.procede(message)
        result_message = await self.resolve(message)
        return result_message.Binary.full

    async def not_in_cache(self, message, question):
        result = await self.connector.resolve(name=question.qname.label,
                                              type=question.qtype,
                                              cls=question.qclass)
        message.header._aa = 1
        if not result:
            # message.header._rcode = 3
            return message
        if isinstance(result, Message):
            message.add_dict(result.__dict__)
        else:
            message.add_dict(result)

    async def resolve(self, message):
        for question in message.question:
            if False:
                in_cache = await self.connector.cache_connector.Cache.get(name=question.qname.label,
                                                                          type=question.qtype,
                                                                          cls=question.qclass)
                if in_cache:
                    json_ = json.dumps(in_cache)
                    message.from_json(json_)
                    message.header._aa = 1
                    return message
                else:
                    await self.not_in_cache(message, question)
                    await self.connector.cache_connector.Cache.set(name=question.qname.label,
                                                                   type=question.qtype,
                                                                   cls=question.qclass,
                                                                   data=message.to_json())
            else:
                await self.not_in_cache(message, question)
                return message
        return message

    def build_message(self, raw_data):
        message = Message.parse_bytes(raw_data)
        return message



from setuptools import setup, find_packages

setup(
    name='neptune_dns_server',  # How you named your package folder (MyLib)
    packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),  # Chose the same as "name"
    version='2019.10.13.2111',  # Start with a small number and increase it with every change you make
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    long_description='''# Neptune DNS Server
Neptune is asynchronous modular DNS server made for flexibility and easy modifications
Its based on Triton dns client and uses different modules to extend it to full server
## Installation
### From PYPI
```
pip3 install neptune-dns-server
```
### From this repo

```
git clone https://git.best-service.online/yurzs/neptune
cd neptune
python3 setup.py install
```
## Usage 
Please edit config.py file for your configuration. Basic install uses 1.1.1.1 and 8.8.8.8 as a resolvers with local redis cache  
To start server use a simple command
```
import neptune
neptune.start_server()
```

TODO write more docs''',
    long_description_content_type='text/markdown',
    description='Async DNS server',  # Give a short description about your library
    author='Yury (Yurzs)',  # Type in your name
    author_email='dev@best-service.online',  # Type in your E-Mail
    url='https://git.best-service.online/yurzs/neptune',  # Provide either the link to your github or to your website
    keywords=['triton', 'DNS', 'client'],  # Keywords that define your package best
    install_requires=['neptune-cache-redis', 'neptune-dnsoverhttps-protocol', 'neptune-resolver-default', 'neptune-resolver-rest', 'triton-dns-client'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3',  # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.7',
    ],
)